"""Subpackage for function and layer service status and management functionality."""

from .functioning import echo_function_versions  # noqa: F401
from .functioning import get_function_version  # noqa: F401
from .functioning import get_function_versions  # noqa: F401
from .functioning import remove_function_version  # noqa: F401
from .layering import echo_layer_versions  # noqa: F401
from .layering import get_layer_version  # noqa: F401
from .layering import get_layer_versions  # noqa: F401
from .layering import remove_layer_version  # noqa: F401



"""Tail the logs for the selected lambda functions."""

import argparse
import datetime
import json
import time
import typing

import botocore.client

from reviser import definitions
from reviser import interactivity


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Get shell auto-completes for this command."""
    return []


def populate_subparser(parser: argparse.ArgumentParser):
    """Populate parser for this command."""
    return


def run(ex: "interactivity.Execution") -> "interactivity.Execution":
    """Execute a bundle operation on the selected function/layer targets."""
    selected = ex.shell.context.get_selected_targets(ex.shell.selection)
    functions = [
        t for t in selected.targets if t.kind == definitions.TargetType.FUNCTION
    ]
    try:
        print("Tailing lambda logs. Press ctrl + C to stop")
        _log_forever(functions)
    except KeyboardInterrupt:
        # This is the expected way to exit.
        pass
    return ex.finalize(
        status="STOPPED",
        message="Stopped tailing logs.",
        info={},
        echo=True,
    )


def _log_forever(targets: typing.List[definitions.Target]):
    """Tail continuously the given targets without end."""
    tokens: typing.Dict[str, str] = {}
    started_at = time.time()
    has_multiple = len([n for t in targets for n in t.names]) > 1
    while True:
        for t in targets:
            for name in t.names:
                log_group = f"/aws/lambda/{name}"
                client = t.client("logs")
                active = _get_active_log_streams(client, log_group)
                for stream in active:
                    stream_key = f"{log_group}/{stream}"
                    token = tokens.get(stream_key)
                    extra = {"nextToken": token} if token else {}
                    response = client.get_log_events(
                        logGroupName=log_group,
                        logStreamName=stream,
                        startFromHead=True,
                        startTime=int(started_at * 1000),
                        **extra,
                    )
                    _print_log_events(response["events"], name, has_multiple)
                    tokens[stream_key] = response["nextForwardToken"]
        time.sleep(1)
        _escape_hook()


def _is_active(log_stream: dict, cut_off: int) -> bool:
    """
    Determine if the given stream is still active and worth tailing.

    :param log_stream:
        A dictionary returned by `describe_log_streams` describing the
        log_stream under consideration.
    :param cut_off:
        The number of seconds to wait before calling a log_stream inactive.
        If the stream does not have a `lastIngestionTime` more recent than
        `cut_off` seconds ago it will be considered not active.
    :return:
        Whether the given log stream is active.
    """
    last_ingest = log_stream["lastIngestionTime"] / 1000
    return time.time() - last_ingest < cut_off


def _print_log_events(
    events: typing.List[dict],
    function_name: str,
    show_name: bool,
):
    """
    Print out the given set of events from a `get_log_events` call.

    :param events:
        The raw events to print out.
    :param function_name:
        The name of the function from which the events came.
    :param show_name:
        Whether to include the function name when printing out the log
        message.
    """
    base_prefix = f"[{function_name}]" if show_name else ""
    for event in events:
        timestamp = datetime.datetime.fromtimestamp(event["timestamp"] / 1000)
        prefix = f"{base_prefix}{timestamp.isoformat()} : "
        try:
            record = json.loads(event["message"])
            message = json.dumps(record, indent=2)
        except json.decoder.JSONDecodeError:
            message = event["message"]
        message = message.strip()
        for line in message.split("\n"):
            print(f"{prefix}{line}")
            prefix = " " * len(prefix)


def _get_active_log_streams(
    client: botocore.client.BaseClient,
    log_group: str,
    cut_off: int = 900,
) -> typing.List[str]:
    """
    Get the active streams within the specified log group.

    :param client:
        The client with which to query cloudwatch.
    :param log_group:
        The log group to search.
    :param cut_off:
        The number of seconds to wait before calling a log_stream inactive.
        If the stream does not have a `lastIngestionTime` more recent than
        `cut_off` seconds ago it will be considered not active.
    """
    response = client.describe_log_streams(
        logGroupName=log_group,
        orderBy="LastEventTime",
        descending=True,
    )
    return [
        log_stream["logStreamName"]
        for log_stream in response["logStreams"]
        if _is_active(log_stream, cut_off)
    ]


def _escape_hook():
    """Escape function for giving unit tests a means to escape the infinite loop."""
    pass



"""Data structure and IO module for lambda layer definitions."""

import dataclasses
import typing
import fnmatch

from reviser.definitions import abstracts
from reviser.definitions import configurations


@dataclasses.dataclass(frozen=True)
class AttachedLayer(abstracts.Specification):
    """Data structure for layer attachments to function targets."""

    target: "configurations.Target"

    @property
    def name(self) -> str:
        """Get the name of the layer attachment."""
        value = self.get("name", default=self.get("arn", default=""))
        if not value.startswith("arn:aws:lambda:"):
            return value.split(":")[0]
        return value.split(":")[6]

    @property
    def version(self) -> typing.Optional[int]:
        """Get the explicit version of the layer for the attachment."""
        if (version := self.get("version")) is not None:
            return int(version)

        value = self.get("arn", default="")
        try:
            return int(value.split(":")[-1])
        except ValueError:
            pass

        value = self.get("name", default="")
        try:
            return int(value.split(":")[-1])
        except ValueError:
            return None

    @property
    def arn(self) -> str:
        """Get the fully-qualified ARN for the layer attachment."""
        if (arn := self.get("arn")) is not None:
            return arn

        value: str = self.get("name", default="")
        if value.startswith("arn:aws:lambda:"):
            return value

        account_id = self.connection.aws_account_id
        region = self.target.aws_region
        name = self.name
        arn = f"arn:aws:lambda:{region}:{account_id}:layer:{name}"
        if (version := self.version) is not None:
            return f"{arn}:{version}"
        return arn

    @property
    def restrictions(self) -> typing.List[str]:
        """List functions to attach this layer to or empty for attach to all."""
        return self.get_as_list("only", default=[]) or []

    @property
    def exclusions(self) -> typing.List[str]:
        """List functions not to attach this layer to."""
        return self.get_as_list("except", default=[]) or []

    def is_attachable(self, function_name: str) -> bool:
        """Specify whether to attach this layer to the specified function."""
        includer = (
            True
            for pattern in self.restrictions
            if fnmatch.fnmatch(function_name, pattern)
        )
        if not next(includer, not self.restrictions):
            return False

        excluder = (False for e in self.exclusions if fnmatch.fnmatch(function_name, e))
        return next(excluder, True)

    def serialize(self) -> dict:
        """Serialize for logging and debugging purposes."""
        return {
            "name": self.name,
            "version": self.version,
            "arn": self.arn,
            "only": self.restrictions,
            "except": self.exclusions,
            "functions": [n for n in self.target.names if self.is_attachable(n)],
        }



"""List versions of the specified lambda targets with info about each version."""

import typing
from argparse import ArgumentParser

from reviser import interactivity
from reviser import servicer


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Get shell auto-completes for this command."""
    return []


def populate_subparser(parser: ArgumentParser):
    """Configure the command arguments for this command."""
    pass


def run(ex: "interactivity.Execution") -> "interactivity.Execution":
    """Execute the listing command."""
    selected = ex.shell.context.get_selected_targets(ex.shell.selection)

    for target in selected.function_targets:
        client = target.client("lambda")
        servicer.echo_function_versions(client, target.names)

    for target in selected.layer_targets:
        client = target.client("lambda")
        function_names = [
            name
            for t in selected.function_targets
            for name in t.names
            if t.aws_region == target.aws_region
        ]
        servicer.echo_layer_versions(client, target.names, function_names)

    return ex.finalize(
        status="LISTED",
        message="Items have been listed.",
        echo=False,
    )



"""Definitions subpackage containing shared data structures used throughout reviser."""

from .abstracts import DataWrapper  # noqa: F401
from .abstracts import Specification  # noqa: F401
from .aws import AwsConnection  # noqa: F401
from .configurations import Bundle  # noqa: F401
from .configurations import Configuration  # noqa: F401
from .configurations import Dependency  # noqa: F401
from .configurations import EnvironmentVariable  # noqa: F401
from .configurations import PipDependency  # noqa: F401
from .configurations import PipperDependency  # noqa: F401
from .configurations import PoetryDependency  # noqa: F401
from .configurations import Target  # noqa: F401
from .contexts import Context  # noqa: F401
from .contexts import SelectedTargets  # noqa: F401
from .enumerations import DefaultFile  # noqa: F401
from .enumerations import DependencyType  # noqa: F401
from .enumerations import RUNTIME_VERSION  # noqa: F401
from .enumerations import TargetType  # noqa: F401
from .infomatics import LambdaFunction  # noqa: F401
from .infomatics import LambdaLayer  # noqa: F401
from .infomatics import LambdaLayerReference  # noqa: F401
from .infomatics import PublishedLayer  # noqa: F401
from .selections import Selection  # noqa: F401



"""Reload the lambda.yaml configuration file from disk."""

import argparse
import typing

import yaml

from reviser import definitions
from reviser import interactivity


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Get shell auto-completes for this command."""
    return []


def populate_subparser(parser: argparse.ArgumentParser):
    """Populate parser for this command."""
    pass


def _merge_target_uuids(
    before: "definitions.Target",
    after: "definitions.Target",
):
    """Match UUIDs from the before to the after targets/children for continuity."""
    after.data["_uuid"] = before.uuid
    after.bundle.data["_uuid"] = before.bundle.uuid
    after.image.data["_uuid"] = before.image.uuid

    mapping = {
        f"{d.kind}:{p}": d
        for d in before.dependencies.sources
        for p in d.get_package_names()
    }

    for dependency in after.dependencies.sources:
        keys = [f"{dependency.kind}:{p}" for p in dependency.get_package_names()]
        match: typing.Optional[definitions.Dependency] = next(
            (d for k, d in mapping.items() if k in keys), None
        )
        if match:
            dependency.data["_uuid"] = match.uuid


def _merge_uuids(
    before: "definitions.Configuration",
    after: "definitions.Configuration",
):
    """Add matching UUIDs from before to the after for continuity."""
    after.data["_uuid"] = before.uuid

    mapping = {name: target for target in before.targets for name in target.names}

    for target in after.targets:
        match: typing.Optional[definitions.Target] = next(
            (t for n, t in mapping.items() if n in target.names),
            None,
        )
        if match:
            _merge_target_uuids(match, target)


def run(ex: "interactivity.Execution") -> "interactivity.Execution":
    """Execute a bundle operation on the selected function/layer targets."""
    before = ex.shell.context
    directory = before.configuration.directory
    connection = before.connection
    after = definitions.Context.load_from_file(
        arguments=before.arguments,
        path=str(directory),
        connection=connection,
    )
    _merge_uuids(before.configuration, after.configuration)
    ex.shell.context = after

    print(yaml.safe_dump(ex.shell.context.configuration.serialize()))
    return ex.finalize(
        status="SUCCESS",
        message="Configuration file has been reloaded from disk.",
        echo=True,
        data={"before": before, "after": after},
    )



"""Subpackage for lambda function management and status functionality."""

import dataclasses
import datetime
import typing
from collections import defaultdict

from botocore.client import BaseClient

from reviser import definitions
from reviser import templating


def get_function_versions(
    lambda_client: BaseClient,
    function_name: str,
) -> typing.List["definitions.LambdaFunction"]:
    """List all of a function's versions and their associated aliases."""
    try:
        paginator = lambda_client.get_paginator("list_versions_by_function")
        versions = [
            definitions.LambdaFunction(item)
            for response in paginator.paginate(FunctionName=function_name)
            for item in response["Versions"]
        ]
    except lambda_client.exceptions.ResourceNotFoundException:
        return []

    aliases = lambda_client.list_aliases(FunctionName=function_name, MaxItems=500)
    aliased = defaultdict(list)
    for a in aliases["Aliases"]:
        aliased[a["FunctionVersion"]].append(a)

    versions.sort(
        key=lambda x: int(1e8 if x.version == "$LATEST" else (x.version or 0)),
    )
    return [
        dataclasses.replace(v, alias_responses=aliased[v.version]) for v in versions
    ]


def remove_function_version(lambda_client, version_arn: str):
    """
    Remove the lambda function version with the specified version ARN.

    This will fail for any versions attached to an alias.

    :param lambda_client:
        Boto3 client for interacting with the lambda function.
    :param version_arn:
        The ARN for the version of the lambda function that should be removed.
    """
    try:
        lambda_client.delete_function(FunctionName=version_arn)
        print("[PRUNED]: {}".format(version_arn))
    except Exception as error:
        print(
            '[FAILED]: Version {} could not be deleted "{}"'.format(version_arn, error)
        )


def echo_function_versions(client: BaseClient, names: typing.List[str]):
    """Display function version info for each specified lambda function."""
    for name in names:
        versions = get_function_versions(client, name)
        if not versions:
            print(f'\n[IGNORED]: No function "{name}" was found.')
            continue

        groups = defaultdict(list)
        for v in versions:
            dt = datetime.datetime.fromisoformat((v.modified or "").split("+")[0])
            groups[str(dt.date())].append({"time": dt.strftime("%H:%M"), "data": v})

        templating.printer(
            "servicer/functioning/echo_versions.jinja2",
            name=name,
            groups=list(sorted(groups.items(), key=lambda x: x[0])),
        )


def get_function_version(
    lambda_client: BaseClient,
    function_name: str,
    qualifier: str = "$LATEST",
) -> "definitions.LambdaFunction":
    """Retrieve the configuration for the specified lambda function."""
    response = lambda_client.get_function_configuration(
        FunctionName=function_name,
        Qualifier=qualifier or "$LATEST",
    )

    request = {
        "FunctionName": function_name,
        "MaxItems": 500,
    }
    if (version := response.get("version")) and version != "$LATEST":
        request["FunctionVersion"] = version
    alias_response = lambda_client.list_aliases(**request)

    return definitions.LambdaFunction(
        response,
        alias_response.get("Aliases") or [],
    )



"""Abstract classes utilized elsewhere for parent inheritance."""

import dataclasses
import pathlib
import typing
import uuid

from reviser.definitions import aws


@dataclasses.dataclass(frozen=True)
class DataWrapper:
    """Base class for data loaded and stored in a dynamic dictionary."""

    #: Data that represents the dynamic representation that is being
    #: wrapped by this object.
    data: dict

    def get(self, *args: str, default: typing.Any = None) -> typing.Any:
        """Fetch the value from the data dictionary."""
        value = self.data or {}
        for k in args:
            if k not in value:
                return default
            value = value[k]
        return value

    def get_first(
        self,
        *args: typing.Iterable[str],
        default: typing.Any = None,
    ) -> typing.Any:
        """
        Fetch the value from the data dictionary in the key-grouped order.

        The default value will be returned instead if none of the key-group args exist
        in the data dictionary.
        """
        found_args = next((a for a in args if self.has(*a)), None)
        if found_args is None:
            return default
        return self.get(*found_args, default=default)

    def get_as_list(
        self,
        *args: str,
        default: typing.Any = None,
    ) -> typing.List[typing.Any]:
        """
        Fetch the value from the data dictionary as a list.

        None values will be returned as an empty list. Values that are
        not lists will be wrapped into single-length lists when returned.
        If the nested key does not exist, the default will be returned
        instead, but if the key is explicitly defined as `null` an empty
        list is returned instead fo the default.
        """
        if self.has(*args):
            value = self.get(*args, default=default)
        else:
            value = default

        if value is None:
            return []

        if isinstance(value, list):
            return value
        return [value]

    def get_first_as_list(
        self,
        *args: typing.Iterable[str],
        default: typing.Any = None,
    ) -> typing.Any:
        """
        Fetch the value from the data dictionary in the key-grouped order.

        The default value will be returned if none of the key-group args
        exist in the data dictionary. If a match is found it will be returned
        as a list in the same fashion as the get_as_list function.
        """
        found_args = next((a for a in args if self.has(*a)), None)
        if found_args is None:
            return default
        return self.get_as_list(*found_args, default=default)

    def get_as_dict(
        self,
        *args: str,
        default: typing.Any = None,
    ) -> typing.Dict[typing.Any, typing.Any]:
        """
        Fetch the value from the data dictionary as a dict.

        None values will be returned as an empty dict. Values that are
        not dicts will raise an error. If the nested key does not exist,
        the default will be returned instead, but if the key is explicitly defined as
        `null` an empty dict is returned instead fo the default.
        """
        if self.has(*args):
            value = self.get(*args, default=default)
        else:
            value = default

        if value is None:
            return {}

        if not isinstance(value, dict):
            raise ValueError(f"{value} must bve of type dict")
        return value

    def get_first_as_dict(
        self,
        *args: typing.Iterable[str],
        default: typing.Any = None,
    ) -> typing.Any:
        """
        Fetch the value from the data dictionary in the key-grouped order.

        The default value will be returned if none of the key-group args
        exist in the data dictionary. If a match is found it will be returned
        as a dict in the same fashion as the get_as_dict function.
        """
        found_args = next((a for a in args if self.has(*a)), None)
        if found_args is None:
            return default
        return self.get_as_dict(*found_args, default=default)

    def has(self, *args: str) -> bool:
        """Determine whether or not the nested key exists."""
        value = self.data or {}
        for k in args:
            if k not in value:
                return False
            value = value[k]
        return True

    def update(self, *args, value: typing.Any = None, overwrite: bool = False):
        """Update the data object if the key is not found to be set already."""
        container = self.data or {}
        for k in args[:-1]:
            if k not in container:
                container[k] = {}
            container = container[k]

        if overwrite or args[-1] not in container:
            container[args[-1]] = value


@dataclasses.dataclass(frozen=True)
class Specification(DataWrapper):
    """Base class for data loaded from the configuration."""

    #: Root directory in which the project being operated on resides.
    directory: pathlib.Path
    #: Specification data loaded from a configuration file that represents
    #: the data for this object.
    data: dict
    #: Connection object that represents the AWS session and account data
    #: associated with the context in which this configuration resides.
    connection: "aws.AwsConnection"

    def __post_init__(self):
        """Add an internal uuid to the data dictionary if one is not already set."""
        if not self.has("_uuid"):
            self.update("_uuid", value=str(uuid.uuid4()), overwrite=False)

    @property
    def uuid(self) -> str:
        """Identify this instance uniquely within the broader configuration."""
        return self.get("_uuid")

    def serialize(self) -> dict:
        """Serialize the object for output representation."""
        return {}



"""
Display help information on the commands available within the shell.

Additional help on each command can be found using the --help flag on the command in
question.
"""

import argparse
import typing
import re

from reviser import commands
from reviser import interactivity
from reviser import templating

HEADER_REGEX = re.compile(r"#+\s+")


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Get shell auto-completes for this command."""
    return []


def populate_subparser(parser: argparse.ArgumentParser):
    """Populate parser for this command."""
    pass


def _get_command_header_doc(name: str, docs: str) -> str:
    """Get the first paragraph of the command documentation."""
    lines = [
        line.rstrip()
        for line in (docs or "").split("\n")
        if not HEADER_REGEX.match(line)
    ]
    start_index = next((i for i, line in enumerate(lines) if line), 0)

    try:
        end_index = lines.index("", start_index) + 1
    except ValueError:
        end_index = len(lines) + 1

    aliases = ""
    if items := commands.REVERSED_ALIASES.get(name):
        aliases = "({}) ".format(", ".join(sorted(items)))

    return "{}{}".format(aliases, " ".join(lines[start_index:end_index]))


def run(ex: "interactivity.Execution") -> "interactivity.Execution":
    """Display basic command help for the available shell commands."""
    command_docs = {
        name: _get_command_header_doc(name, docs)
        for name, command_module in commands.COMMANDS.items()
        if (docs := command_module.__doc__)
    }
    templating.printer(
        "commands/helper/help.jinja2",
        commands=list(sorted(command_docs.items(), key=lambda x: x[0])),
    )
    return ex.finalize(status="HELPED", message="Help displayed.")



"""
Upload the bundled contents to the upload S3 bucket and then publish a new version.

This will be carried out for each of the lambda targets with that new bundle and
any modified settings between the current configuration and that target's
existing configuration. This command will fail if a target being deployed
has not already been bundled.
"""

import argparse
import os
import string
import typing

from reviser import deploying
from reviser import interactivity


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Get shell auto-completes for this command."""
    return ["--description", "--dry-run"]


def populate_subparser(parser: argparse.ArgumentParser):
    """Populate parser for this command."""
    parser.add_argument(
        "--description",
        help="""
        Specify a message to assign to the version published by
        the deploy command.
        """,
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="""
        If set, the deploy operation will be exercised without
        actually carrying out the actions. This can be useful to
        validate the deploy process without side effects.
        """,
    )


def run(ex: "interactivity.Execution"):
    """Execute a bundle operation on the selected function/layer targets."""
    description = string.Template(ex.args.get("description") or "").substitute(
        os.environ
    )
    print("\n\n")
    deployed_targets = deploying.deploy(
        context=ex.shell.context,
        selection=ex.shell.selection,
        description=description,
        dry_run=ex.args.get("dry_run") or False,
    )
    print("\n")
    return ex.finalize(
        status="DEPLOYED",
        message="Selected items have been deployed.",
        info={
            "dry_run": ex.args.get("dry_run"),
            "items": [n for t in deployed_targets for n in t.names],
            "description": description,
        },
        echo=True,
    )



"""Image configuration data structures and IO module."""

import dataclasses
import typing

from reviser.definitions import abstracts
from reviser.definitions import configurations
from reviser.definitions import enumerations


@dataclasses.dataclass(frozen=True)
class Image(abstracts.Specification):
    """Define the function image config that will be generated."""

    target: "configurations.Target"

    @property
    def configured(self) -> bool:
        """Determine if URIs are configured."""
        return (
            bool(self.get("uri"))
            and self.target.kind == enumerations.TargetType.FUNCTION
        )

    def get_region_uri(self, region: str) -> typing.Optional[str]:
        """Get the uri of the image to use in the lambda."""
        uri = self.get("uri")
        if isinstance(uri, dict):
            return uri.get(region)
        return uri

    @property
    def entrypoint(self) -> typing.Optional[typing.List[str]]:
        """
        Get the custom entrypoint of the image.

        If not specified the default of the image will be used.
        """
        custom_entrypoint = self.get("entrypoint")
        if isinstance(custom_entrypoint, str):
            custom_entrypoint = [custom_entrypoint]
        return custom_entrypoint if self.configured is not None else None

    @property
    def cmd(self) -> typing.Optional[typing.List[str]]:
        """
        Get the custom command of the image.

        If not specified the default of the image will be used.
        """
        custom_cmd = self.get("cmd")
        if isinstance(custom_cmd, str):
            custom_cmd = [custom_cmd]
        return custom_cmd if self.configured is not None else None

    @property
    def workingdir(self) -> typing.Optional[str]:
        """
        Get the custom workdir of the image.

        If not specified the default of the image will be used.
        """
        return self.get("workingdir") if self.configured is not None else None

    def serialize(self) -> dict:
        """Serialize the object for output representation."""
        return {
            "uri_map": self.get("uri"),
            "entrypoint": self.entrypoint,
            "cmd": self.cmd,
            "workingdir": self.workingdir,
        }
